/*
 * gsampler-device-manager.c
 * Copyright (C) Lilian Monchalin 2011 <isaya07@users.sf.net>
 * 
 * gsampler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * gsampler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <glib/gi18n-lib.h>

#include "gsampler-device-manager.h"
#include "gsampler-client-device.h"
#include "gsampler-window.h"
#include "gsampler-utils.h"
#include "gsampler-device.h"
#include "gsampler-option.h"

G_DEFINE_TYPE (GsamplerDeviceManager, gsampler_device_manager, GTK_TYPE_BOX);

static GsamplerDeviceManager* gsampler_device_manager__instance = NULL;
static gboolean gsampler_device_manager_status = FALSE;
static GSList *audio_device_list = NULL;
static GSList *midi_device_list = NULL;

static gboolean 
gsampler_device_manager_create_device_form (GsamplerDeviceManager* self, gint type, gint id) 
{
	GSList *list = NULL;
	GsamplerDevice* device;
	g_return_val_if_fail (self != NULL, FALSE);
	
	if (type == 0) 
	{
		list = audio_device_list;
	} else if (type == 1) 
	{
		list = midi_device_list;
	} else 
	{
		return FALSE;
	}
	if (g_id_list_exist (list, id)) 
	{
		return TRUE;
	}
	device = gsampler_device_new (type);
	if (!gsampler_device_active (device, id)) 
	{
		gsampler_main_window_send_message ("warning", "Cant get audio device info");
		gtk_widget_destroy ((GtkWidget*) device);
		return FALSE;
	} else 
	{
		if (type == 0) 
		{
			audio_device_list = g_id_list_add (audio_device_list, id, g_object_ref (device));
			gtk_box_pack_start (self->priv->audio_box, (GtkWidget*) device, FALSE, FALSE, 0);
		} else if (type == 1) 
		{
			midi_device_list = g_id_list_add (midi_device_list, id, g_object_ref (device));
			gtk_box_pack_start (self->priv->midi_box, (GtkWidget*) device, FALSE, FALSE, 0);
		}
		gtk_widget_show ((GtkWidget*) device);
	}
	return TRUE;
}

static void 
gsampler_device_manager_create_device (GsamplerDeviceManager* self, gint type) 
{
	GtkDialog* dialog = NULL;
	GtkComboBox* driver_name = NULL;
	GtkEntry* device_name = NULL;
	gchar** drivers = NULL;
	gint driver_lenght = 0;
	gchar* defaults = NULL;
	gint id = -1;
	GsamplerClient* client = NULL;
	GsamplerOption* option = NULL;
	GsamplerMainWindow* win = NULL;
	GtkLabel* driver_label;
	GtkLabel* name_label;
	GtkHBox* hbox1;
	GtkHBox* hbox2;
	GtkBox* vbox1;
	gint y;
	gint result;
	g_return_if_fail (self != NULL);
	
	client = gsampler_client_get_instance ();
	option = gsampler_option_get_instance ();
	if (type == 0) 
	{
		drivers = gsampler_client_get_audio_driver_list (client);
		defaults = gsampler_option_get_string (option, "Client", "default-audio");
	} else if (type == 1) 
	{
		drivers = gsampler_client_get_midi_driver_list (client);
		defaults = gsampler_option_get_string (option, "Client", "default-midi");
	} else 
	{
		return;
	}
	win = gsampler_main_window_get_instance ();
	dialog = (GtkDialog*) gtk_dialog_new_with_buttons (_("Create Device"), (GtkWindow*) win, 
	                                      			   GTK_DIALOG_DESTROY_WITH_PARENT, 
	                                      			   GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
	                                      			   NULL);
	GtkWidget* apply = gtk_dialog_add_button (dialog, GTK_STOCK_APPLY, GTK_RESPONSE_APPLY);
	driver_name = gsampler_text_combo_create ();
	device_name = (GtkEntry*) gtk_entry_new ();
	driver_label = (GtkLabel*) gtk_label_new (_("Select Driver : "));
	name_label = (GtkLabel*) gtk_label_new (_("Set the Name : "));
	hbox1 = (GtkHBox*) gtk_hbox_new (TRUE, 5);
	hbox2 = (GtkHBox*) gtk_hbox_new (TRUE, 5);
	gtk_box_pack_start ((GtkBox*) hbox1, (GtkWidget*) driver_label, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) hbox1, (GtkWidget*) driver_name, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) hbox2, (GtkWidget*) name_label, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) hbox2, (GtkWidget*) device_name, FALSE, TRUE, 0);
	vbox1 = (GtkBox*) gtk_dialog_get_content_area (dialog);
	gtk_box_pack_start (vbox1, GTK_WIDGET (hbox1), FALSE, TRUE, 0);
	gtk_box_pack_start (vbox1, GTK_WIDGET (hbox2), FALSE, TRUE, 0);
	gtk_widget_show_all ((GtkWidget*) dialog);

	y = 0;	
	int driver_it;
	if (drivers != NULL)
		driver_lenght = g_strv_length (drivers);
	for (driver_it = 0; driver_it < driver_lenght; driver_it = driver_it + 1) 
	{
		gchar* driver;
		driver = g_strdup (drivers[driver_it]);
		gsampler_text_combo_append (driver_name, driver);
		if (g_strcmp0 (driver, defaults) == 0) 
		{
			gtk_combo_box_set_active ((GtkComboBox*) driver_name, y);
		}
		y++;
		g_free (driver);
	}
	
	if (type == 0) 
	{
		id = gsampler_client_get_audio_output_count (client);
	} else if (type == 1) 
	{
		id = gsampler_client_get_midi_input_count (client);
	}
	gtk_entry_set_text (device_name, g_strdup_printf ("Gsampler %i", id + 1));
	gtk_entry_set_activates_default (device_name, TRUE);
	gtk_dialog_set_default_response (dialog, GTK_RESPONSE_APPLY);
	g_object_set (apply, "has-focus", TRUE, NULL);
	result = gtk_dialog_run (dialog);
	if (result == GTK_RESPONSE_APPLY) 
	{
		gchar* driver;
		gchar* tmpname;

		driver = gsampler_text_combo_get_active (driver_name);
		tmpname = g_strdup (gtk_entry_get_text (device_name));
		id = -1;
		if (type == 0) 
		{
			id = gsampler_client_create_audio_device (client, driver, tmpname);
		} else if (type == 1) 
		{
			id = gsampler_client_create_midi_device (client, driver, tmpname);
		}
		g_free (tmpname);
		g_free (driver);
	}
	if (id == (-1))
	{
		gsampler_main_window_send_message ("warning", _("Can't create audio device"));
	}
	gtk_widget_destroy ((GtkWidget*) dialog);
	g_object_unref (win);
	g_object_unref (option);
	g_object_unref (client);
	g_free (defaults);
	g_strfreev (drivers);
}

static void 
gsampler_device_manager_delete_device_form (GsamplerDeviceManager* self, gint type, gint id) 
{
	GSList *list = NULL;
	GsamplerDevice* device;
	g_return_if_fail (self != NULL);
	
	if (type == 0) 
	{
		list = audio_device_list;
	} else if (type == 1) 
	{
		list = midi_device_list;
	} else 
	{
		return;
	}
	device = (GsamplerDevice*) g_id_list_get (list, id);
	if (GSAMPLER_IS_DEVICE (device)) 
	{
		gtk_widget_destroy ((GtkWidget*) device);
		if (type == 0) 
		{
			audio_device_list = g_id_list_remove (list, id);
		} else if (type == 1) 
		{
			midi_device_list = g_id_list_remove (list, id);
		}
	} else 
	{
		g_warning ("Can't supress device form invalid list id");
	}
}

static void 
gsampler_device_manager_on_audio_add (GtkButton* sender, gpointer data) 
{
	GsamplerDeviceManager* self;
	self = GSAMPLER_DEVICE_MANAGER (data);
	g_return_if_fail (GSAMPLER_IS_DEVICE_MANAGER (self));
	g_return_if_fail (sender != NULL);
	
	if (gsampler_device_manager_status == TRUE) 
	{
		gsampler_device_manager_create_device (self, 0);
	}
}

static void 
gsampler_device_manager_on_midi_add (GtkButton* sender, gpointer data) 
{
	GsamplerDeviceManager* self;
	self = GSAMPLER_DEVICE_MANAGER (data);
	g_return_if_fail (GSAMPLER_IS_DEVICE_MANAGER (self));
	g_return_if_fail (sender != NULL);
	
	if (gsampler_device_manager_status == TRUE) 
	{
		gsampler_device_manager_create_device (self, 1);
	}
}

static void 
gsampler_device_manager_on_info_event (GsamplerClient* sender, gint type, gint id, gpointer data) 
{
	GSList* list = NULL;
	GsamplerDevice* device;
	GsamplerDeviceManager* self;
	self = GSAMPLER_DEVICE_MANAGER (data);
	g_return_if_fail (GSAMPLER_IS_DEVICE_MANAGER (self));
	
	if (type == 0) 
	{
		list = audio_device_list;
	} else if (type == 1) 
	{
		list = midi_device_list;
	} else 
	{
		return;
	}
	
	device = (GsamplerDevice*) g_id_list_get (list, id);
	if (device != NULL) 
	{
		gsampler_device_update_info (device);
		//g_object_unref (device);
	} else 
	{
		g_warning ("Can't update device info invalid list id device don't exsist");
	}
}

static void 
gsampler_device_manager_on_update_event (GsamplerClient* sender, gchar** devlist,
                                         gint type, gpointer data) 
{
	GSList *list = NULL;
	gint devlist_length = 0;
	GsamplerDeviceManager* self = GSAMPLER_DEVICE_MANAGER (data);
	g_return_if_fail (GSAMPLER_IS_DEVICE_MANAGER (self));
	
	while (!gsampler_device_manager_status) 
	{
		g_usleep ((guint) 300000);
	}
	gsampler_device_manager_status = FALSE;
	if (type == 0) 
	{
		list = audio_device_list;
	} else if (type == 1)
	{
		list = midi_device_list;
	} else 
	{
		return;
	}
	if (devlist != NULL)
		devlist_length = g_strv_length (devlist);
	while (list) 
	{
		gint id, dev_it;
		gboolean result = TRUE;
		id = ((struct Struct_list_widget*) (list->data))->Sid;
		for (dev_it = 0; dev_it < devlist_length; dev_it = dev_it + 1) 
		{
			gchar* dev;
			dev = g_strdup (devlist[dev_it]);
			if (g_strcmp0 (dev, "NONE") == 0) 
			{
				result = TRUE;
			} else {
				gint new_id;
				new_id = atoi (dev);
				if (id == new_id) 
				{
					result = FALSE;
				}
			}
			g_free (dev);			
		}
		if (result) 
		{
			gsampler_device_manager_delete_device_form (self, type, id);
			break;
		}
		list = g_slist_next(list);
	}
	gint dev_it;
	for (dev_it = 0; dev_it < devlist_length; dev_it = dev_it + 1) 
	{
		gchar* dev;
		dev = g_strdup (devlist[dev_it]);
		if (g_strcmp0 (dev, "NONE") != 0) 
		{
			gint id;
			id = atoi (dev);
			if (!g_id_list_exist (list, id)) 
			{
				if (!gsampler_device_manager_create_device_form (self, type, id)) 
				{
					g_warning ("Cant't create device form %d", id);
				}
			}
		}
		g_free (dev);
	}
	gsampler_device_manager_status = TRUE;
}

static void 
gsampler_device_manager_init (GsamplerDeviceManager * self) 
{
	self->priv = GSAMPLER_DEVICE_MANAGER_GET_PRIVATE (self);
}

static GObject * 
gsampler_device_manager_constructor (GType type, guint n_construct_properties, 
                               		 GObjectConstructParam * construct_properties) 
{
	GObject * obj;
	GsamplerDeviceManager * self = NULL;
	GtkBox* hbox1 = NULL;
	GtkImage* audio_img = NULL;
	GtkLabel* audio_lbl = NULL;
	GtkButton* audio_button = NULL;
	GtkScrolledWindow* audio_scrol = NULL;
	GtkBox* hbox2 = NULL;
	GtkImage* midi_img = NULL;
	GtkLabel* midi_lbl = NULL;
	GtkButton* midi_button = NULL;
	GtkScrolledWindow* midi_scrol = NULL;
											
	obj = G_OBJECT_CLASS (gsampler_device_manager_parent_class)->constructor (type, n_construct_properties, construct_properties);
	self = GSAMPLER_DEVICE_MANAGER (obj);
	gtk_container_set_border_width ((GtkContainer*) self, 2);
	gtk_orientable_set_orientation ((GtkOrientable*) self, GTK_ORIENTATION_VERTICAL);
	hbox1 = (GtkBox*) gtk_hbox_new (FALSE, 5);
#if GTK_CHECK_VERSION (2, 90, 0)
	gtk_widget_set_margin_bottom ((GtkWidget*) hbox1, 3);
#endif
	audio_img = (GtkImage*) gtk_image_new_from_stock (GSAMPLER_STOCK_CARD, GTK_ICON_SIZE_BUTTON);
	audio_lbl = (GtkLabel*) gtk_label_new (NULL);
	gtk_label_set_markup (audio_lbl, _("<b>Audio Device :</b>"));
	audio_button = (GtkButton*) gtk_button_new_from_stock (GTK_STOCK_ADD);
	audio_scrol = (GtkScrolledWindow*) gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (audio_scrol, GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type (audio_scrol, GTK_SHADOW_IN);
	GtkWidget *viewport1 = gtk_viewport_new (gtk_scrolled_window_get_hadjustment (audio_scrol),
	                                         gtk_scrolled_window_get_vadjustment (audio_scrol));
	gtk_viewport_set_shadow_type ((GtkViewport *) viewport1, GTK_SHADOW_NONE);
	self->priv->audio_box = (GtkBox*) gtk_vbox_new (FALSE, 0);
	hbox2 = (GtkBox*) gtk_hbox_new (FALSE, 5);
#if GTK_CHECK_VERSION (2, 90, 0)
	gtk_widget_set_margin_bottom ((GtkWidget*) hbox2, 3);
	gtk_widget_set_margin_top ((GtkWidget*) hbox2, 3);
#endif
	midi_img = (GtkImage*) gtk_image_new_from_stock (GSAMPLER_STOCK_CARD, GTK_ICON_SIZE_BUTTON);
	midi_lbl = (GtkLabel*) gtk_label_new (NULL);
	gtk_label_set_markup (midi_lbl, _("<b>Midi Device :</b>"));
	midi_button = (GtkButton*) gtk_button_new_from_stock (GTK_STOCK_ADD);
	midi_scrol = (GtkScrolledWindow*) gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (midi_scrol, GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type (midi_scrol, GTK_SHADOW_IN);
	GtkWidget *viewport2 = gtk_viewport_new (gtk_scrolled_window_get_hadjustment (midi_scrol),
	                                         gtk_scrolled_window_get_vadjustment (midi_scrol));
	gtk_viewport_set_shadow_type ((GtkViewport *) viewport2, GTK_SHADOW_NONE);
	self->priv->midi_box = (GtkBox*) gtk_vbox_new (FALSE, 0);
	
	gtk_box_pack_start (hbox1, GTK_WIDGET (audio_img), FALSE, TRUE, (guint) 5);
	gtk_box_pack_start (hbox1, GTK_WIDGET (audio_lbl), TRUE, TRUE, (guint) 5);
	gtk_box_pack_start (hbox1, GTK_WIDGET (audio_button), FALSE, TRUE, (guint) 5);
	gtk_container_add((GtkContainer*) audio_scrol, viewport1);
	gtk_container_add ((GtkContainer*) viewport1, (GtkWidget*) self->priv->audio_box);
	gtk_box_pack_start (hbox2, GTK_WIDGET (midi_img), FALSE, TRUE, (guint) 5);
	gtk_box_pack_start (hbox2, GTK_WIDGET (midi_lbl), TRUE, TRUE, (guint) 5);
	gtk_box_pack_start (hbox2, GTK_WIDGET (midi_button), FALSE, TRUE, (guint) 5);
	gtk_container_add((GtkContainer*) midi_scrol, viewport2);
	gtk_container_add ((GtkContainer*) viewport2, (GtkWidget*) self->priv->midi_box);
	gtk_box_pack_start ((GtkBox*) self, GTK_WIDGET (hbox1), FALSE, TRUE, 3);
	gtk_box_pack_start ((GtkBox*) self, GTK_WIDGET (audio_scrol), TRUE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) self, GTK_WIDGET (hbox2), FALSE, TRUE, 3);
	gtk_box_pack_start ((GtkBox*) self, GTK_WIDGET (midi_scrol), TRUE, TRUE, 0);
											
	g_signal_connect_object (audio_button, "clicked", 
	                         (GCallback) gsampler_device_manager_on_audio_add, self, 0);
	g_signal_connect_object (midi_button, "clicked", 
	                         (GCallback) gsampler_device_manager_on_midi_add, self, 0);
	gtk_widget_show_all ((GtkWidget*) self);
	gtk_widget_set_sensitive ((GtkWidget*) self, FALSE);
	return obj;
}

static void 
gsampler_device_manager_finalize (GObject* obj) 
{
	GsamplerDeviceManager * self;
	self = GSAMPLER_DEVICE_MANAGER (obj);
	g_object_unref (self->priv->audio_box);
	g_object_unref (self->priv->midi_box);
	G_OBJECT_CLASS (gsampler_device_manager_parent_class)->finalize (obj);
}

static void 
gsampler_device_manager_class_init (GsamplerDeviceManagerClass * klass) 
{
	gsampler_device_manager_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (GsamplerDeviceManagerPrivate));
	G_OBJECT_CLASS (klass)->constructor = gsampler_device_manager_constructor;
	G_OBJECT_CLASS (klass)->finalize = gsampler_device_manager_finalize;
	g_signal_new ("device_event", 
	              GSAMPLER_TYPE_DEVICE_MANAGER, 
	              G_SIGNAL_RUN_LAST,
	              0, NULL, NULL, 
	              g_cclosure_marshal_VOID__STRING,
	              G_TYPE_NONE, 
	              1, G_TYPE_STRING);
}

GsamplerDeviceManager* 
gsampler_device_manager_new (void) 
{
	return g_object_new (GSAMPLER_TYPE_DEVICE_MANAGER, NULL);
}

GsamplerDeviceManager* 
gsampler_device_manager_get_instance (void) 
{
	if (gsampler_device_manager__instance == NULL) 
	{
		gsampler_device_manager__instance = gsampler_device_manager_new ();
	}
	return g_object_ref (gsampler_device_manager__instance);
}

GSList * 
gsampler_device_manager_get_audios (void) 
{
	return audio_device_list;
}


GSList * 
gsampler_device_manager_get_midis (void) 
{
	return midi_device_list;
}

void 
gsampler_device_manager_active (GsamplerDeviceManager* self) 
{
	GsamplerClient* client = NULL;
	gchar** dev_list;
	g_return_if_fail (self != NULL);
	
	client = gsampler_client_get_instance ();
	gsampler_device_manager_status = TRUE;
	dev_list = gsampler_client_get_available_audio_output (client);
	gsampler_device_manager_on_update_event (client, dev_list, 0, self);
	g_strfreev (dev_list);
	dev_list = gsampler_client_get_available_midi_input (client);
	gsampler_device_manager_on_update_event (client, dev_list, 1, self);
	g_strfreev (dev_list);
	g_signal_connect_object (client, "device-count-event", 
	                         (GCallback) gsampler_device_manager_on_update_event, self, 0);
	g_signal_connect_object (client, "device-info-event", 
	                         (GCallback) gsampler_device_manager_on_info_event, self, 0);
	gtk_widget_set_sensitive ((GtkWidget*) self, TRUE);
	g_object_unref (client);
}


void 
gsampler_device_manager_inactive (GsamplerDeviceManager* self) 
{
	GsamplerClient* client = NULL;
	g_return_if_fail (self != NULL);
	
	client = gsampler_client_get_instance ();
	gsampler_device_manager_status = FALSE;
	g_signal_handlers_disconnect_by_func (client, 
	                                      gsampler_device_manager_on_update_event, self);
	g_signal_handlers_disconnect_by_func (client, 
	                                      gsampler_device_manager_on_info_event, self);
	gtk_widget_set_sensitive ((GtkWidget*) self, FALSE);
	g_object_unref (client);
}
