/*
 * gsampler-keyboard.c
 * Copyright (C) Lilian Monchalin 2011 <isaya07@users.sf.net>
 * 
 * gsampler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * gsampler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n-lib.h>

#include "gsampler-keyboard.h"
#include "gsampler-channel-manager.h"
#include "gsampler-channel.h"
#include "gsampler-utils.h"

G_DEFINE_TYPE (GsamplerKeyboard, gsampler_keyboard, GTK_TYPE_BOX);

#define KEYBOARD_HEIGHT 70

static void 
gsampler_keyboard_draw_digit (GtkWidget* widget, cairo_t* cr, gint key, gint w) {
	PangoRectangle log_rect = {0};
	PangoLayout* layout;
	gchar* markup;
	gdouble text_w;
	gdouble text_h;
	gdouble x;
	g_return_if_fail (widget != NULL);
	g_return_if_fail (cr != NULL);
	
	layout = pango_cairo_create_layout (cr);
	markup = g_strdup_printf ("<span size=\"8000\">%d</span>", (key / 12) - 2);
	pango_layout_set_markup (layout, markup, strlen (markup));
	pango_layout_get_extents (layout, NULL, &log_rect);
	text_w = (gdouble) (log_rect.width / PANGO_SCALE);
	text_h = (gdouble) (log_rect.height / PANGO_SCALE);
	x = (w * (key + 0.75)) / 128.0;
	cairo_set_source_rgb (cr, (gdouble) 0, (gdouble) 0, (gdouble) 0);
	cairo_move_to (cr, (x - (text_w / 2)) + 1, (KEYBOARD_HEIGHT - text_h) + 0.5);
	pango_cairo_update_layout (cr, layout);
	pango_cairo_show_layout (cr, layout);
	g_free (markup);
	g_object_unref (layout);
}

static void 
gsampler_keyboard_draw_key (GtkWidget* widget, gint key, gboolean state) {
	GtkAllocation alloc = {0};
	GdkColor color = {0};
	cairo_t* cr;
	gint h;
	gint w;
	gdouble bh;
	gint note;
	gint x;
	gint x2;
	gint x3;
	gint x4;
	gint w1;
	g_return_if_fail (widget != NULL);
	
	cr = gdk_cairo_create (gtk_widget_get_window (widget));
	gtk_widget_get_allocation (widget, &alloc);
	h = KEYBOARD_HEIGHT;
	w = (gint) alloc.width;
	bh = h * 0.6;
	note = (gint) ((key + 3) % 12);
	x = (gint) (((w * key) / 128.0) + 1.5);
	x2 = (gint) ((w * (key + 1.5)) / 128.0);
	x3 = (gint) ((w * (key + 1)) / 128.0);
	x4 = (gint) (((w * (key - 0.5)) / 128.0) + 2);
	w1 = x3 - x;
	if (state) {
		gdk_color_parse ("red", &color);
	} else if (note == 1 || note == 4 || note == 6 || note == 9 || note == 11) {
		gdk_color_parse ("black", &color);
	} else {
		gdk_color_parse ("white", &color);
	}
	cairo_set_source_rgb (cr, (gdouble) color.red, (gdouble) color.green, (gdouble) color.blue);
	switch (note) {
		case 0: case 5: case 10:
			cairo_rectangle (cr, (gdouble) x, (gdouble) 1, w1 + 0.5, bh);
			cairo_fill (cr);
			cairo_rectangle (cr, (gdouble) x4, bh + 1, (gdouble) (x2 - x4), (h - bh) - 2);
			cairo_fill (cr);
			break;
		case 2: case 7:
			cairo_rectangle (cr, (gdouble) x, (gdouble) 1, (gdouble) w1, bh);
			cairo_fill (cr);
			cairo_rectangle (cr, (gdouble) x4, bh + 1, (gdouble) (x3 - x4), (h - bh) - 2);
			cairo_fill (cr);
			break;
		case 3: case 8:
			cairo_rectangle (cr, (gdouble) x, (gdouble) 1, w1 + 0.5, bh);
			cairo_fill (cr);
			cairo_rectangle (cr, (gdouble) x, bh + 1, (gdouble) (x2 - x), (h - bh) - 2);
			cairo_fill (cr);
			if (note == 3) {
				gsampler_keyboard_draw_digit (widget, cr, key, w);
			}
			break;
		default:
			cairo_rectangle (cr, (gdouble) x, (gdouble) 1, (gdouble) (w1 + 1), bh - 1);
			cairo_fill (cr);
			break;
	}
	cairo_destroy (cr);
}

#if GTK_CHECK_VERSION (2, 90, 0)
static gboolean 
gsampler_keyboard_on_draw_event (GtkWidget* widget, cairo_t* cr, gpointer data)
{
#else	
static gboolean
gsampler_keyboard_on_expose_event (GtkWidget* widget, GdkEventExpose* event, gpointer data) 
{
	cairo_t* cr = gdk_cairo_create (gtk_widget_get_window(widget));
#endif
	GdkColor white = {0};
	GdkColor black = {0};
	GtkAllocation alloc = {0};
	gint h;
	gint w;
	gint i;
	gdouble bh;
	g_return_val_if_fail (cr != NULL, FALSE);
	
	gdk_color_parse ("white", &white);
	gdk_color_parse ("black", &black);
	gtk_widget_get_allocation (widget, &alloc);
	h = KEYBOARD_HEIGHT;
	w = (gint) alloc.width;
	bh = h * 0.6;
	cairo_set_source_rgb (cr, white.red, white.green, white.blue);
	cairo_rectangle (cr, 0.5, 0.5, (gdouble) (w - 1), (gdouble) (h - 1));
	cairo_fill (cr);
	cairo_set_source_rgb (cr, black.red, black.green, black.blue);
	for (i = 0; i < 128; i++) {
		gint note;
		gint x;
		note = (gint) ((i + 3) % 12);
		x = (gint) ((w * i) / 128.0);
		if (note == 1 || note == 4 || note == 6 || note == 9 || note == 11) {
			gint x1;
			gint x2;
			x1 = (gint) (((w * (i + 0.5)) / 128.0) + 0.5);
			cairo_move_to (cr, x1 + 0.5, bh + 0.5);
			cairo_line_to (cr, x1 + 0.5, (gdouble) (h - 1));
			cairo_stroke (cr);
			x2 = (gint) (((w * (i + 1)) / 128.0) + 0.5);
			cairo_rectangle (cr, x, 1, (gdouble) ((x2 - x) + 1), bh);
			cairo_fill (cr);
		} else if (note == 3 || note == 8) {
			cairo_move_to (cr, x + 0.5, (gdouble) 1);
			cairo_line_to (cr, x + 0.5, (gdouble) (h - 1));
			cairo_stroke (cr);
			if (note == 3) gsampler_keyboard_draw_digit (widget, cr, i, w);
		}
	}
	return TRUE;
}

static void
gsampler_keyboard_send_midi_note (GsamplerKeyboard* self, int pitch, int vel, gboolean type) {
	gint chan;
	g_return_if_fail (self != NULL);

	chan = gtk_combo_box_get_active ((GtkComboBox*) self->priv->combo);
	if (type) {
		self->priv->active_key = pitch;
		g_signal_emit_by_name (self, "note-on", chan, pitch, vel);
		gsampler_keyboard_draw_key ((GtkWidget*) self->priv->draw_area, pitch, TRUE);
	} else {
		g_signal_emit_by_name (self, "note-off", chan, pitch, vel);
		gsampler_keyboard_draw_key ((GtkWidget*) self->priv->draw_area, pitch, FALSE);
		self->priv->active_key = 0;
	}
}

static gboolean 
gsampler_keyboard_on_button_pressed (GtkWidget* sender, GdkEventButton* event, gpointer data) {
	GtkAllocation alloc = {0};
	gint vel = 0;
	GsamplerKeyboard* self = (GsamplerKeyboard*) data;
	g_return_val_if_fail (self != NULL, FALSE);
	
	if (event->button == 1) {
		gint pitch;
		gint note;
		gtk_widget_get_allocation ((GtkWidget*) self->priv->draw_area, &alloc);
		pitch = (((gint) event->x) * 128) / alloc.width;
		note = (gint) ((pitch + 3) % 12);
		if (note == 1 || note == 4 || note == 6 || note == 9 || note == 11) {
			vel = (gint) ((event->y * 127) / (KEYBOARD_HEIGHT * 0.6));
		} else {
			vel = (gint) ((event->y * 127) / KEYBOARD_HEIGHT);
		}
		gsampler_keyboard_send_midi_note (self, pitch, vel, TRUE);
	}
	return FALSE;
}

static gboolean 
gsampler_keyboard_on_button_released (GtkWidget* sender, GdkEventButton* event, gpointer data) {
	GtkAllocation alloc = {0};
	gint vel = 0;
	GsamplerKeyboard* self = (GsamplerKeyboard*) data;
	g_return_val_if_fail (self != NULL, FALSE);

	if (event->button == 1) {
		gint pitch;
		gint note;
		gtk_widget_get_allocation ((GtkWidget*) self->priv->draw_area, &alloc);
		pitch = (gint) ((event->x * 128) / alloc.width);
		note = (gint) ((pitch + 3) % 12);
		if (note == 1 || note == 4 || note == 6 || note == 9 || note == 11) {
			vel = (gint) ((event->y * 127) / (KEYBOARD_HEIGHT * 0.6));
		} else {
			vel = (gint) ((event->y * 127) / KEYBOARD_HEIGHT);
		}
		gsampler_keyboard_send_midi_note (self, pitch, vel, FALSE);
	}
	return FALSE;
}

static gboolean
gsampler_keyboard_on_motion_notify (GtkWidget* sender, GdkEventButton* event, gpointer data) {
	GtkAllocation alloc = {0};
	gint pitch;
	GsamplerKeyboard* self = (GsamplerKeyboard*) data;
	g_return_val_if_fail (self != NULL, FALSE);
	
	if (self->priv->active_key > 0 && event->y >= 0 && event->y < KEYBOARD_HEIGHT) {
		gtk_widget_get_allocation ((GtkWidget*) self->priv->draw_area, &alloc);
		pitch = (gint) ((event->x * 128) / alloc.width);
		if (self->priv->active_key != pitch) {
			int vel = 0;
			int note;
			note = (int) ((pitch + 3) % 12);
			if (note == 1 || note == 4 || note == 6 || note == 9 || note == 11) {
				vel = (gint) ((event->y * 127) / (KEYBOARD_HEIGHT * 0.6));
			} else {
				vel = (gint) ((event->y * 127) / KEYBOARD_HEIGHT);
			}
			gsampler_keyboard_send_midi_note (self, self->priv->active_key, vel, FALSE);
			self->priv->active_key = pitch;
			gsampler_keyboard_send_midi_note (self, pitch, vel, TRUE);
		}
	}
	return FALSE;
}

static gboolean
gsampler_keyboard_on_leave_notify (GtkWidget* sender, GdkEventButton* event, gpointer data) {
	GsamplerKeyboard* self;
	self = (GsamplerKeyboard*) data;
	g_return_val_if_fail (self != NULL, FALSE);

	if (self->priv->active_key != 0)
		gsampler_keyboard_send_midi_note (self, self->priv->active_key, 0, FALSE);
	return FALSE;
}

static void 
gsampler_keyboard_on_channels_updated (GsamplerChannelManager* sender, gpointer data) {
	GSList* list = NULL;
	GsamplerKeyboard* self = (GsamplerKeyboard*) data;
	g_return_if_fail (self != NULL);

	gsampler_text_combo_remove_all (self->priv->combo);
	list = gsampler_channel_manager_get_channels ();
	while (list) {
		GsamplerChannel* channel;
		gchar* chan_text;

		channel = (GsamplerChannel*) ((struct Struct_list_widget*) (list->data))->Sdata;
		chan_text = g_strdup_printf ("%d - %s", channel->priv->info->id, channel->priv->info->inst_name);
		gsampler_text_combo_insert (self->priv->combo, chan_text, channel->priv->info->id);
		g_free (chan_text);
		//g_object_unref (channel);
		list = g_slist_next (list);
	}

	gtk_combo_box_set_active ((GtkComboBox*) self->priv->combo, 0);
}

static void
gsampler_keyboard_on_scroll_realize (GtkWidget *widget, gpointer data)
{
	g_return_if_fail (widget != NULL);

	GtkAdjustment *hadj = gtk_scrolled_window_get_hadjustment ((GtkScrolledWindow*) widget);
	gdouble lower = gtk_adjustment_get_lower (hadj);
	gdouble upper = gtk_adjustment_get_upper (hadj);
	gdouble value = (upper-lower)/8;
	gtk_adjustment_set_value (hadj, value);
}

static void 
g_cclosure_user_marshal_VOID__INT_INT_INT (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data) {
	typedef void (*GMarshalFunc_VOID__INT_INT_INT) (gpointer data1, gint arg_1, gint arg_2, gint arg_3, gpointer data2);
	register GMarshalFunc_VOID__INT_INT_INT callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = (GCClosure *) closure;
	g_return_if_fail (n_param_values == 4);
	if (G_CCLOSURE_SWAP_DATA (closure)) {
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else {
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = (GMarshalFunc_VOID__INT_INT_INT) (marshal_data ? marshal_data : cc->callback);
	callback (data1, g_value_get_int (param_values + 1), g_value_get_int (param_values + 2), g_value_get_int (param_values + 3), data2);
}

static void 
gsampler_keyboard_init (GsamplerKeyboard * keyboard) {	
	keyboard->priv = GSAMPLER_KEYBOARD_GET_PRIVATE (keyboard);

	keyboard->priv->active_key = 0;
}

static GObject * 
gsampler_keyboard_constructor (GType type, guint n_construct_properties, 
                               GObjectConstructParam * construct_properties) 
{
	GObject * obj;
	GtkScrolledWindow *scroll = NULL;
	GsamplerKeyboard *keyboard = NULL;
	GtkBox* hbox = NULL;
	GtkLabel* lbl = NULL;
	
	obj = G_OBJECT_CLASS (gsampler_keyboard_parent_class)->constructor (type, n_construct_properties, construct_properties);
	keyboard = GSAMPLER_KEYBOARD (obj);
	gtk_container_set_border_width ((GtkContainer*) keyboard, 2);
	gtk_orientable_set_orientation ((GtkOrientable*) keyboard, GTK_ORIENTATION_VERTICAL);
	lbl = (GtkLabel*) gtk_label_new (NULL);
	gtk_label_set_markup (lbl, _("<b>Channel : </b>"));
	keyboard->priv->combo = gsampler_text_combo_create ();
	g_object_set ((GtkWidget*) keyboard->priv->combo, "width-request", 200, NULL);
	hbox = (GtkBox*) gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (hbox, (GtkWidget*) lbl, FALSE, TRUE, 0);
	gtk_box_pack_start (hbox, (GtkWidget*) keyboard->priv->combo, FALSE, TRUE, 0);
#if GTK_CHECK_VERSION (2, 90, 0)
	gtk_widget_set_halign ((GtkWidget*) hbox, GTK_ALIGN_CENTER);
#endif
	keyboard->priv->draw_area = (GtkDrawingArea*) gtk_drawing_area_new ();
	g_object_set ((GtkWidget*) keyboard->priv->draw_area, "width-request", 896, NULL);
	g_object_set ((GtkWidget*) keyboard->priv->draw_area, "height-request", KEYBOARD_HEIGHT, NULL);
	scroll = (GtkScrolledWindow*) gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (scroll, GTK_POLICY_AUTOMATIC, GTK_POLICY_NEVER);

	gtk_scrolled_window_set_shadow_type (scroll, GTK_SHADOW_ETCHED_OUT);
	gtk_scrolled_window_add_with_viewport (scroll, (GtkWidget*) keyboard->priv->draw_area);
	gtk_box_pack_start ((GtkBox*) keyboard, (GtkWidget*) hbox, FALSE, TRUE, (guint) 6);
	gtk_box_pack_start ((GtkBox*) keyboard, (GtkWidget*) scroll, FALSE, TRUE, (guint) 0);
	gtk_widget_add_events ((GtkWidget*) keyboard->priv->draw_area, 
	                       (GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_LEAVE_NOTIFY_MASK |
               			   GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK));
#if GTK_CHECK_VERSION (2, 90, 0)
	g_signal_connect_object ((GtkWidget*) keyboard->priv->draw_area, "draw", 
	                         (GCallback) gsampler_keyboard_on_draw_event, keyboard, 0);
#else
	g_signal_connect_object ((GtkWidget*) keyboard->priv->draw_area, "expose_event", 
	                         (GCallback) gsampler_keyboard_on_expose_event, keyboard, 0);
#endif
	gtk_widget_show_all ((GtkWidget*) keyboard);
	gtk_widget_set_sensitive ((GtkWidget*) keyboard, FALSE);
	g_signal_connect (scroll, "realize", 
	                  (GCallback) gsampler_keyboard_on_scroll_realize, NULL);
	return obj;
}

static void 
gsampler_keyboard_finalize (GObject* obj) 
{
	G_OBJECT_CLASS (gsampler_keyboard_parent_class)->finalize (obj);
}

static void 
gsampler_keyboard_class_init (GsamplerKeyboardClass * klass) 
{
	gsampler_keyboard_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (GsamplerKeyboardPrivate));
	G_OBJECT_CLASS (klass)->constructor = gsampler_keyboard_constructor;
	G_OBJECT_CLASS (klass)->finalize = gsampler_keyboard_finalize;
	g_signal_new ("note_on", GSAMPLER_TYPE_KEYBOARD, G_SIGNAL_RUN_LAST, 0, NULL, NULL,
	              g_cclosure_user_marshal_VOID__INT_INT_INT, G_TYPE_NONE, 3,
	              G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);
	g_signal_new ("note_off", GSAMPLER_TYPE_KEYBOARD, G_SIGNAL_RUN_LAST, 0, NULL, NULL,
	              g_cclosure_user_marshal_VOID__INT_INT_INT, G_TYPE_NONE, 3,
	              G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);
}

GsamplerKeyboard* 
gsampler_keyboard_new (void) 
{
	return g_object_new (GSAMPLER_TYPE_KEYBOARD, NULL);
}

void gsampler_keyboard_active (GsamplerKeyboard* keyboard) {
	GsamplerChannelManager* chan_manag;
	g_return_if_fail (keyboard != NULL);
	
	chan_manag = gsampler_channel_manager_get_instance ();
	g_signal_connect_object (chan_manag, "channels-updated", 
	                         (GCallback) gsampler_keyboard_on_channels_updated, keyboard, 0);
	g_signal_connect_object ((GtkWidget*) keyboard->priv->draw_area, "button-press-event", 
	                         (GCallback) gsampler_keyboard_on_button_pressed, keyboard, 0);
	g_signal_connect_object ((GtkWidget*) keyboard->priv->draw_area, "button-release-event", 
	                         (GCallback) gsampler_keyboard_on_button_released, keyboard, 0);
	g_signal_connect_object ((GtkWidget*) keyboard->priv->draw_area, "motion-notify-event", 
	                         (GCallback) gsampler_keyboard_on_motion_notify, keyboard, 0);
	g_signal_connect_object ((GtkWidget*) keyboard->priv->draw_area, "leave-notify-event", 
	                         (GCallback) gsampler_keyboard_on_leave_notify, keyboard, 0);
	gsampler_keyboard_on_channels_updated (chan_manag, keyboard);
	gtk_widget_set_sensitive ((GtkWidget*) keyboard, TRUE);
	g_object_unref (chan_manag);
}


void gsampler_keyboard_inactive (GsamplerKeyboard* keyboard) {
	GsamplerChannelManager* chan_manag;
	g_return_if_fail (keyboard != NULL);

	chan_manag = gsampler_channel_manager_get_instance ();
	g_signal_handlers_disconnect_by_func (chan_manag, 
	                                      gsampler_keyboard_on_channels_updated, keyboard);
	g_signal_handlers_disconnect_by_func (keyboard->priv->draw_area, 
	                                      gsampler_keyboard_on_button_pressed, keyboard);
	g_signal_handlers_disconnect_by_func (keyboard->priv->draw_area, 
	                                      gsampler_keyboard_on_button_released, keyboard);
	g_signal_handlers_disconnect_by_func (keyboard->priv->draw_area, 
	                                      gsampler_keyboard_on_motion_notify, keyboard);
	g_signal_handlers_disconnect_by_func (keyboard->priv->draw_area, 
	                                      gsampler_keyboard_on_leave_notify, keyboard);
	gtk_widget_set_sensitive ((GtkWidget*) keyboard, FALSE);
	g_object_unref (chan_manag);
}