/*
 * gsampler-window.c
 * Copyright (C) Lilian Monchalin 2011 <isaya07@users.sf.net>
 *
 * gsampler is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * gsampler is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include <string.h>

#include "gsampler-window.h"
#include "gsampler-preference.h"
#include "gsampler-utils.h"
#include "gsampler-assistant.h"

G_DEFINE_TYPE (GsamplerMainWindow, gsampler_main_window, GTK_TYPE_WINDOW);

static GsamplerMainWindow* manwin_instance = NULL;
static gboolean manwin_status = FALSE;

// Pre declaration
static void gsampler_main_window_on_client_error_event (GsamplerClient* client, const gchar* type,
                                            			const gchar* mes, gpointer data);

enum {
	PROP_0,
	PROP_TOTAL_VOICE,
	PROP_TOTAL_STREAM
};

#if GTK_CHECK_VERSION (2,90,0)
#define APP_ICON "gsampler.svg"
#else
#define APP_ICON "gsampler.png"
#endif


static void 
gsampler_main_window_save_state (GsamplerMainWindow* self) 
{
	gint width = 0;
	gint height = 0;
	gint tmp_pos;
	gboolean tmp_vis;
	g_return_if_fail (self != NULL);
	
	gtk_window_get_size ((GtkWindow*) self, &width, &height);
	gsampler_option_set_integer (self->priv->_option, "General", "width", width);
	gsampler_option_set_integer (self->priv->_option, "General", "height", height);
	tmp_pos = gtk_paned_get_position (self->priv->paned1);
	gsampler_option_set_integer (self->priv->_option, "General", "paned1", tmp_pos);
	tmp_pos = gtk_paned_get_position (self->priv->paned2);
	gsampler_option_set_integer (self->priv->_option, "General", "paned2", tmp_pos);
	tmp_vis = gtk_widget_get_visible ((GtkWidget*) self->priv->_toolbar);
	gsampler_option_set_boolean (self->priv->_option, "General", "toolbar", tmp_vis);
	tmp_vis = gtk_widget_get_visible ((GtkWidget*) self->priv->_keyboard);
	gsampler_option_set_boolean (self->priv->_option, "General", "keyboard", tmp_vis);
	tmp_vis = gtk_widget_get_visible ((GtkWidget*) self->priv->_statusbar);
	gsampler_option_set_boolean (self->priv->_option, "General", "statusbar", tmp_vis);
	tmp_vis = gtk_widget_get_visible ((GtkWidget*) self->priv->console);
	gsampler_option_set_boolean (self->priv->_option, "General", "console", tmp_vis);
	tmp_pos = gtk_notebook_get_current_page (self->priv->notebook);
	gsampler_option_set_integer (self->priv->_option, "General", "tab", tmp_pos);
	gsampler_option_save_keyfile (self->priv->_option);
}

static void 
gsampler_main_window_update_title (GsamplerMainWindow* self) 
{
	gchar* name = NULL;
	g_return_if_fail (self != NULL);

	if (self->priv->_session != NULL)
		name = g_strdup (gsampler_session_get_name (self->priv->_session));
	if (name != NULL && strlen (name) != 0) 
	{
		if (gsampler_session_get_is_updated (self->priv->_session)) 
		{
			name = g_strconcat ("*", name, NULL);
		}
		gtk_window_set_title ((GtkWindow*) self, g_strconcat (name, " :: Gsampler", NULL));
		g_free (name);
	} else {
		gtk_window_set_title ((GtkWindow*) self, "Gsampler");
	}
}

static void 
gsampler_main_window_on_session_notify (GObject* object, GParamSpec* pspec, gpointer data) 
{
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_if_fail (GSAMPLER_IS_MAIN_WINDOW (self));
	g_return_if_fail (pspec != NULL);
	
	gsampler_main_window_update_title (self);
}

static gboolean 
gsampler_main_window_on_quit (GtkWidget* sender, GdkEvent* event, gpointer data) 
{
	GsamplerClient* client;
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (sender);
	g_return_val_if_fail (self != NULL, FALSE);

	if (gsampler_session_get_is_updated (self->priv->_session)) 
	{
		GtkWidget* dialog;
		gint response;
		dialog = gtk_message_dialog_new ((GtkWindow*) self, GTK_DIALOG_DESTROY_WITH_PARENT,
		                                 GTK_MESSAGE_WARNING, GTK_BUTTONS_NONE, "%s", _("Warning"));
		g_object_set (dialog, "secondary-text", " all unsaved changes will be lost...", NULL);
		gtk_dialog_add_buttons ((GtkDialog*) dialog, GTK_STOCK_QUIT, GTK_RESPONSE_OK, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_APPLY, NULL, NULL);
		gtk_dialog_set_default_response ((GtkDialog*) dialog, GTK_RESPONSE_APPLY);
		response = gtk_dialog_run ((GtkDialog*) dialog);
		switch (response) 
		{
			case GTK_RESPONSE_CANCEL:
				gtk_widget_destroy (dialog);
				return TRUE;
			case GTK_RESPONSE_APPLY:
				if (!gsampler_session_save_session (self->priv->_session)) {
					gtk_widget_destroy (dialog);
					return TRUE;
				}
				break;
			default:
				break;
		}
		gtk_widget_destroy (dialog);
	}
	gsampler_main_window_save_state (self);
	client = gsampler_client_get_instance ();
	gsampler_client_stop (client, FALSE);
	g_object_unref (client);
	gtk_widget_destroy ((GtkWidget*) self);
	gtk_main_quit ();
	return FALSE;
}

static void 
gsampler_main_window_menu_item_select_cb (GtkMenuItem* item, gpointer data) 
{
	gchar* message = NULL;
	GtkAction* action;
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (item != NULL);

	action = g_object_ref ((GtkAction*) g_object_get_data ((GObject*) item, "gtk-action"));
	g_return_if_fail (action != NULL);
	g_object_get ((GObject*) action, "tooltip", &message, NULL);
	if (message != NULL) 
	{
		gtk_statusbar_push ((GtkStatusbar*) self->priv->_statusbar, 1, message);
	}
	g_object_unref (action);
	g_free (message);
}

static void 
gsampler_main_window_menu_item_deselect_cb (GtkMenuItem* item, gpointer data) 
{
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (item != NULL);
	
	gtk_statusbar_pop ((GtkStatusbar*) self->priv->_statusbar, 1);
}

static void 
gsampler_main_window_connect_proxy_cb (GtkUIManager* uimanager, GtkAction* action,
                                       GtkWidget* proxy, gpointer data) 
{
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (action != NULL);
	g_return_if_fail (proxy != NULL);
	
	if (GTK_IS_MENU_ITEM (proxy)) 
	{
		g_signal_connect_object (proxy, "select", 
		                         (GCallback) gsampler_main_window_menu_item_select_cb, self, 0);
		g_signal_connect_object (proxy, "deselect", 
		                         (GCallback) gsampler_main_window_menu_item_deselect_cb, self, 0);
	}
}

static void 
gsampler_main_window_disconnect_proxy_cb (GtkUIManager* uimanager, GtkAction* action,
                                          GtkWidget* proxy, gpointer data) 
{
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (action != NULL);
	g_return_if_fail (proxy != NULL);
	
	if (GTK_IS_MENU_ITEM (proxy)) 
	{
		g_signal_handlers_disconnect_by_func (proxy, 
	                                      	  gsampler_main_window_menu_item_select_cb, self);
		g_signal_handlers_disconnect_by_func (proxy, 
	                                      	  gsampler_main_window_menu_item_deselect_cb, self);
	}
}

void 
gsampler_main_window_set_status_progress (GsamplerMainWindow* self, const gchar* type, gdouble frac,
                                          const gchar* mes, gpointer data) 
{
	g_return_if_fail (self != NULL);
	g_return_if_fail (type != NULL);
	g_return_if_fail (mes != NULL);
	
	if (g_strcmp0 (type, "error") == 0) 
	{
		gsampler_status_bar_set_progress (self->priv->_statusbar, &frac, "");
		gsampler_main_window_send_message ("warning", mes);
	} else if (g_strcmp0 (type, "prog") == 0) 
	{
		gsampler_status_bar_set_progress (self->priv->_statusbar, &frac, mes);
	} else if (g_strcmp0 (type, "up") == 0) 
	{
		//gsampler_main_window_activate_client (self);
	}
}

void
gsampler_main_window_set_status_pulse (GsamplerMainWindow* self, gboolean watch)
{
	g_return_if_fail (self != NULL);

	gsampler_status_bar_pulse (self->priv->_statusbar, watch);
}

void 
gsampler_main_window_send_message (const gchar* level, const gchar* message) 
{
	GtkMessageType err = 0;
	gchar* first = NULL;
	GtkWidget* dialog = NULL;
	GtkWindow* window = NULL;
	g_return_if_fail (level != NULL);
	g_return_if_fail (message != NULL);

	window = g_object_ref (manwin_instance);
	if (g_strcmp0 (level, "info") == 0) 
	{
		err = GTK_MESSAGE_INFO;
		first = g_strdup (_("Info"));
	} else if (g_strcmp0 (level, "warning") == 0) 
	{
		err = GTK_MESSAGE_WARNING;
		first = g_strdup (_("Warning"));
	} else if (g_strcmp0 (level, "error") == 0) 
	{
		err = GTK_MESSAGE_ERROR;
		first = g_strdup (_("Error"));
	} else 
	{
		err = GTK_MESSAGE_OTHER;
		first = g_strdup (_("Other"));
	}

	dialog = gtk_message_dialog_new (window, GTK_DIALOG_DESTROY_WITH_PARENT,
	                                 err, GTK_BUTTONS_OK, "%s", first);
	g_object_set ((GtkWindow*) dialog, "default-height", 300, NULL);
	g_object_set ((GtkWindow*) dialog, "default-width", 400, NULL);
	g_object_set (dialog, "secondary-text", message, NULL);
	gtk_dialog_run ((GtkDialog*) dialog);
	gtk_widget_destroy (dialog);
	g_object_unref (window);
	g_free (first);
}

static void 
gsampler_main_window_on_volume_changed (GtkAdjustment* adj, gpointer data) 
{
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (adj != NULL);
	
	if (manwin_status) 
	{
		gdouble vol;
		GsamplerClient* client;

		vol = gtk_adjustment_get_value (self->priv->main_vol);
		client = gsampler_client_get_instance ();
		g_return_if_fail (client != NULL);
		if (!gsampler_client_set_master_volume (client, vol)) 
		{
			gsampler_main_window_send_message ("warning", _("Can't set volume"));
		}
		g_object_unref (client);
	}
}

static void 
gsampler_main_window_on_about (GsamplerMainWindow* self) 
{
	GdkPixbuf* pixbuf = NULL;
	GtkAboutDialog* dialog = NULL;
	gchar *load_file = NULL;
	GError * error = NULL;
	g_return_if_fail (self != NULL);

	dialog = (GtkAboutDialog*) gtk_about_dialog_new ();
	load_file = gsampler_utils_get_data_file (APP_ICON);
	if (load_file != NULL)
	{
		pixbuf = gdk_pixbuf_new_from_file_at_scale (load_file, 96, 96, TRUE, &error);
		if (error != NULL) 
		{
			g_warning ("Gsampler error: %s", error->message);
			g_error_free (error);
		} else 
		{
			gtk_about_dialog_set_logo (dialog, pixbuf);
		}
	}
	gtk_about_dialog_set_program_name (dialog, PACKAGE_TARNAME);
	gtk_about_dialog_set_version (dialog, PACKAGE_VERSION);
	gtk_about_dialog_set_comments (dialog, _("Frontend for Linuxsampler"));
	gtk_about_dialog_set_copyright (dialog, "Copyright © Lilian Monchalin");
	gtk_about_dialog_set_license (dialog, "Released under the GNU General Public License V3.\n" \
"Please notice that this is an alpha soft.\n" \
"Please report bugs, comments, suggest to: http://sourceforge.net/p/gsampler/tickets/");
	gtk_about_dialog_set_website (dialog, "http://sourceforge.net/projects/gsampler/");
	gtk_about_dialog_set_website_label (dialog, _("Home page"));
	gtk_window_set_transient_for ((GtkWindow*) dialog, (GtkWindow*) self);
	gtk_dialog_run ((GtkDialog*) dialog);
	gtk_widget_destroy ((GtkWidget*) dialog);
	g_object_unref (pixbuf);
}

static void 
gsampler_main_window_on_total_voice_event (GsamplerClient* sender, gint val, gpointer data) 
{
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_if_fail (self != NULL);

	if (self->priv->total_voice != NULL)
		g_free (self->priv->total_voice);
	self->priv->total_voice = g_strdup_printf ("%i", val);
	gtk_widget_queue_draw ((GtkWidget*) self->priv->master_draw);
}

static void 
gsampler_main_window_on_total_stream_event (GsamplerClient* sender, gint val, gpointer data) 
{
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_if_fail (self != NULL);

	if (self->priv->total_stream != NULL)
		g_free (self->priv->total_stream);
	self->priv->total_stream = g_strdup_printf ("%i", val);
	gtk_widget_queue_draw ((GtkWidget*) self->priv->master_draw);
}

static void 
gsampler_main_window_desactivate_client (GsamplerMainWindow* self) 
{
	GsamplerClient* client = NULL;
	g_return_if_fail (self != NULL);
	
	manwin_status = FALSE;
	client = gsampler_client_get_instance ();
	if (client != NULL) {
		g_signal_handlers_disconnect_by_func (client, 
	                                      	  gsampler_main_window_on_total_stream_event, self);
		g_signal_handlers_disconnect_by_func (client, 
	                                      	  gsampler_main_window_on_total_voice_event, self);
		g_signal_handlers_disconnect_by_func (client, 
	                                      	  gsampler_main_window_on_client_error_event, self);
	}
	gsampler_channel_manager_inactive (self->priv->_chan_manag);
	gsampler_device_manager_inactive (self->priv->_dev_manag);
	gsampler_midi_map_inactive (self->priv->_midi_map);
	gsampler_data_base_inactive (self->priv->_data_base);
	gsampler_keyboard_inactive (self->priv->_keyboard);
	gsampler_console_inactive (self->priv->console);
	GsamplerUiManager* uimanager = gsampler_ui_manager_get_instance ();
	gsampler_ui_manager_group_set_sensitive ((GsamplerUiManager*) uimanager, "client", FALSE);
	g_object_unref (uimanager);
	g_signal_handlers_disconnect_by_func (self->priv->main_vol, 
	                                      gsampler_main_window_on_volume_changed, self);
	g_signal_handlers_disconnect_by_func ((GObject*) self->priv->_session, 
	                                      gsampler_main_window_on_session_notify, self);
	g_signal_handlers_disconnect_by_func (self->priv->_keyboard, 
	                                      gsampler_channel_manager_on_note_on, self);
	g_signal_handlers_disconnect_by_func (self->priv->_keyboard, 
	                                      gsampler_channel_manager_on_note_off, self);
	g_object_unref (client);
	self->priv->status = FALSE;
}

static void 
gsampler_main_window_activate_client (GsamplerMainWindow* self) 
{
	GsamplerClient* client = NULL;
	gdouble vol = 0;
	gdouble tmpvol;
	g_return_if_fail (self != NULL);
	
	manwin_status = TRUE;
	client = gsampler_client_get_instance ();
	gsampler_channel_manager_active (self->priv->_chan_manag);
	gsampler_device_manager_active (self->priv->_dev_manag);
	gsampler_midi_map_active (self->priv->_midi_map);
	gsampler_keyboard_active (self->priv->_keyboard);
	gsampler_console_active (self->priv->console);
	gsampler_data_base_active (self->priv->_data_base);
	GsamplerUiManager* uimanager = gsampler_ui_manager_get_instance ();
	gsampler_ui_manager_group_set_sensitive ((GsamplerUiManager*) uimanager, "client", TRUE);
	g_object_unref (uimanager);
	tmpvol = double_parse (gsampler_client_get_master_volume (client));
	vol = tmpvol * 100;
	gtk_adjustment_set_value (self->priv->main_vol, vol);
	g_signal_connect_object (self->priv->main_vol, "value-changed", 
	                         (GCallback) gsampler_main_window_on_volume_changed, self, 0);
	g_signal_connect_object (client, "total-stream-event", 
	                         (GCallback) gsampler_main_window_on_total_stream_event, self, 0);
	g_signal_connect_object (client, "total-voice-event", 
	                         (GCallback) gsampler_main_window_on_total_voice_event, self, 0);
	g_signal_connect_object (client, "client-serveur-error-event", 
	                         (GCallback) gsampler_main_window_on_client_error_event, self, 0);
	g_signal_connect_object ((GObject*) self->priv->_session, "notify", 
	                         (GCallback) gsampler_main_window_on_session_notify, self, 0);
	g_signal_connect_object (self->priv->_keyboard, "note-on", 
	                         (GCallback) gsampler_channel_manager_on_note_on, self->priv->_chan_manag, 0);
	g_signal_connect_object (self->priv->_keyboard, "note-off", 
	                         (GCallback) gsampler_channel_manager_on_note_off, self->priv->_chan_manag, 0);
	g_object_unref (client);
	self->priv->status = TRUE;
}

static void 
gsampler_main_window_on_client_error_event (GsamplerClient* client, const gchar* type,
                                            const gchar* mes, gpointer data) 
{
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (mes != NULL);

	if (g_strcmp0 (type, "COM") == 0) {
		if (!gtk_widget_get_visible ((GtkWidget*) self->priv->_statusbar)) {
			gtk_widget_show ((GtkWidget*) self->priv->_statusbar);
		}
		gsampler_status_bar_set_error_message (self->priv->_statusbar, mes);
	} else if (g_strcmp0 (type, "CRASH") == 0) {
		gsampler_main_window_desactivate_client (self);	
		gsampler_main_window_send_message ("warning", mes);
	}
}

static void 
gsampler_main_window_on_new_session (GsamplerMainWindow* self, gboolean reset) 
{
	GsamplerClient* client;
	g_return_if_fail (self != NULL);

	client = gsampler_client_get_instance ();
	if (gsampler_client_client_reset (client))
	{
		if (reset) {
			gsampler_main_window_set_wait_cursor (TRUE);
			gsampler_main_window_desactivate_client (self);
			if (gsampler_client_reset_client (client))
				gsampler_main_window_activate_client (self);
			gsampler_main_window_set_wait_cursor (FALSE);
		}
	}
	g_object_unref (client);
	gsampler_session_clear (self->priv->_session);
}

static void
gsampler_main_window_restart_server (GsamplerMainWindow* self)
{
	GsamplerClient* client;
	g_return_if_fail (self != NULL);
	
	gsampler_main_window_set_wait_cursor (TRUE);
	client = gsampler_client_get_instance ();
	if (self->priv->status)
		gsampler_main_window_desactivate_client (self);
	if (gsampler_client_restart (client))
		gsampler_main_window_activate_client (self);
	g_object_unref (client);
	gsampler_main_window_set_wait_cursor (FALSE);
}

static void 
gsampler_main_window_on_preferences (GsamplerMainWindow* self) 
{
	GsamplerPreference* pref;
	g_return_if_fail (self != NULL);

	pref = gsampler_preference_new ();
	gtk_dialog_run ((GtkDialog*) pref);
}

static void 
gsampler_main_window_on_action_activated (GtkUIManager* uimanager, GtkAction* action, gpointer data) 
{
	const gchar* name = NULL;
	GQuark Action;
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_if_fail (self != NULL);
	g_return_if_fail (action != NULL);
	
	name = gtk_action_get_name (action);
	Action = (name == NULL) ? 0 : g_quark_from_string (name);

	if (Action == g_quark_from_string ("FileNew")) 
	{
		if (gsampler_session_ask_overide (self->priv->_session)) 
		{
			gsampler_main_window_on_new_session (self, FALSE);
		}
	} else if (Action == g_quark_from_string ("FileOpen")) 
	{
		if (gsampler_session_ask_overide (self->priv->_session)) 
		{
			gsampler_main_window_on_new_session (self, TRUE);
			gsampler_session_open_session (self->priv->_session);
		}
	} else if (Action == g_quark_from_string ("FileSave")) 
	{
		gsampler_session_save_session (self->priv->_session);
	} else if (Action == g_quark_from_string ("FileSaveAs")) 
	{
		gsampler_session_save_session_as (self->priv->_session);
	} else if (Action == g_quark_from_string ("FileQuit")) 
	{
		gsampler_main_window_on_quit ((GtkWidget*) self, NULL, NULL);
	} else if (Action == g_quark_from_string ("EditPreferences")) {
		gsampler_main_window_on_preferences (self);
	} else if (Action == g_quark_from_string ("ViewToolbar")) 
	{
		GtkToggleAction* toggle_action;
		toggle_action = (GtkToggleAction*) action;
		if (gtk_toggle_action_get_active (toggle_action)) 
		{
			gtk_widget_show ((GtkWidget*) self->priv->_toolbar);
		} else 
		{
			gtk_widget_hide ((GtkWidget*) self->priv->_toolbar);
		}
	} else if (Action == g_quark_from_string ("ViewStatusbar")) 
	{
		GtkToggleAction* toggle_action;
		toggle_action = (GtkToggleAction*) action;
		if (gtk_toggle_action_get_active (toggle_action)) 
		{
			gtk_widget_show ((GtkWidget*) self->priv->_statusbar);
#if GTK_CHECK_VERSION (2, 90, 0)
			gtk_window_set_has_resize_grip ((GtkWindow*) self, TRUE);
#else
			gtk_statusbar_set_has_resize_grip ((GtkStatusbar*) self->priv->_statusbar, TRUE);
#endif
		} else 
		{
			gtk_widget_hide ((GtkWidget*) self->priv->_statusbar);
#if GTK_CHECK_VERSION (2, 90, 0)
			gtk_window_set_has_resize_grip ((GtkWindow*) self, FALSE);
#else
			gtk_statusbar_set_has_resize_grip ((GtkStatusbar*) self->priv->_statusbar, FALSE);
#endif
		}
	} else if (Action == g_quark_from_string ("ViewKeyboard")) 
	{
		GtkToggleAction* toggle_action;
		toggle_action = (GtkToggleAction*) action;
		if (gtk_toggle_action_get_active (toggle_action)) 
		{
			gtk_widget_show ((GtkWidget*) self->priv->_keyboard);
		} else 
		{
			gtk_widget_hide ((GtkWidget*) self->priv->_keyboard);
		}
	} else if (Action == g_quark_from_string ("ViewConsole")) 
	{
		GtkToggleAction* toggle_action;
		toggle_action = (GtkToggleAction*) action;
		if (gtk_toggle_action_get_active (toggle_action)) 
		{
			gtk_widget_show ((GtkWidget*) self->priv->console);
		} else 
		{
			gtk_widget_hide ((GtkWidget*) self->priv->console);
		}
	} else if (Action == g_quark_from_string ("ServeurReset")) 
	{
		if (gsampler_session_ask_overide (self->priv->_session)) 
		{
			gsampler_main_window_on_new_session (self, TRUE);
		}
	} else if (Action == g_quark_from_string ("ServeurConnect")) 
	{
		if (gsampler_session_ask_overide (self->priv->_session)) 
		{
			gsampler_main_window_restart_server (self);
		}
	} else if (Action == g_quark_from_string ("StartAssistant")) 
	{
		GsamplerAssistant *assist = gsampler_assistant_new ();
		gtk_widget_show_all ((GtkWidget *) assist);
		g_signal_connect_swapped (assist, "close",
			                   	  (GCallback) gtk_widget_destroy, assist);
	} else if (Action == g_quark_from_string ("HelpAbout")) 
	{
		gsampler_main_window_on_about (self);		
	}
}

#if GTK_CHECK_VERSION (2, 90, 0)
static gboolean 
gsampler_main_window_on_draw_event (GtkWidget* sender, cairo_t* context, gpointer data) 
{
#else	
static gboolean
gsampler_main_window_on_expose_event (GtkWidget* sender, GdkEventExpose* event, gpointer data) 
{
	cairo_t* context = gdk_cairo_create (gtk_widget_get_window(sender));
#endif
	GdkColor red = {0};
	GdkColor green = {0};
	GtkAllocation alloc = {0};
	gdouble xc;
	cairo_pattern_t* pattern;
	PangoFontDescription* font_description;
	PangoLayout* layout;
	gint width = 0;
	gint height = 0;
	gint place1;
	gint place2;
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_val_if_fail (self != NULL, FALSE);
	g_return_val_if_fail (context != NULL, FALSE);

	gdk_color_parse ("red", &red);
	gdk_color_parse ("green", &green);
	gtk_widget_get_allocation ((GtkWidget*) sender, &alloc);
	xc = (gdouble) ((0 + alloc.width) / 2);
	cairo_set_source_rgba (context, 0, 0, 0, 1);
	cairo_rectangle (context, 0, 0, alloc.width, alloc.height);
	cairo_fill (context);
	pattern = cairo_pattern_create_linear (xc, 0, xc, alloc.height);
	cairo_pattern_add_color_stop_rgba (pattern, 0, 1, 1, 1, 0);
	cairo_pattern_add_color_stop_rgba (pattern, 0.4, 1, 1, 1, 0.2);
	cairo_pattern_add_color_stop_rgba (pattern, 0.6, 1, 1, 1, 0.2);
	cairo_pattern_add_color_stop_rgba (pattern, 1, 1, 1, 1, 0);
	cairo_set_source (context, pattern);
	cairo_rectangle (context, 0, 0, alloc.width, alloc.height);
	cairo_fill (context);
	font_description = pango_font_description_from_string (self->priv->font);
	layout = pango_cairo_create_layout (context);
	pango_layout_set_font_description (layout, font_description);
	cairo_set_source_rgb (context, (gdouble) red.red, (gdouble) red.green, (gdouble) red.blue);
	pango_layout_set_text (layout, g_strdup_printf ("%s/%s",self->priv->total_stream, self->priv->total_voice), -1);
	pango_layout_get_size (layout, &width, &height);
	place1 = (alloc.height - (height / PANGO_SCALE)) / 2;
	place2 = (alloc.width - (width / PANGO_SCALE)) / 2;
	cairo_move_to (context, (gdouble) place2, (gdouble) place1);
	pango_cairo_update_layout (context, layout);
	pango_cairo_show_layout (context, layout);
	g_object_unref (layout);
	pango_font_description_free (font_description);
	cairo_pattern_destroy (pattern);

	return TRUE;
}

static GObject* 
gsampler_main_window_constructor (GType type, 
                                  guint n_construct_params, 
                                  GObjectConstructParam *construct_params) 
{
	GObject * obj;
	GsamplerMainWindow * self = NULL;
	GtkWidget* main_box = NULL;
	GtkWidget* sec_box = NULL;
	GtkWidget* vol_box = NULL;
	GtkWidget* vol_img = NULL;
	GtkWidget* vol_lbl = NULL;
	GtkScale* vol_scale = NULL;
	GtkWidget* map_label = NULL;
	GtkWidget* db_label = NULL;
	
	obj = G_OBJECT_CLASS (gsampler_main_window_parent_class)->constructor (type, n_construct_params,
	                                                                       construct_params);
	self = GSAMPLER_MAIN_WINDOW (obj);
	self->priv->_option = gsampler_option_get_instance ();
	g_object_set ((GtkWindow*) self, "default-height", 
	              gsampler_option_get_integer (self->priv->_option, "General", "height"), NULL);
	g_object_set ((GtkWindow*) self, "default-width",
	              gsampler_option_get_integer (self->priv->_option, "General", "width"), NULL);
	gtk_window_set_icon_name ((GtkWindow*) self, PACKAGE_NAME);
	self->priv->_ui_manag = gsampler_ui_manager_get_instance ();
	gtk_window_add_accel_group ((GtkWindow*) self,
	                            gtk_ui_manager_get_accel_group ((GtkUIManager*) self->priv->_ui_manag));
								   
	g_signal_connect_object ((GtkUIManager*) self->priv->_ui_manag, "connect-proxy",
	                         (GCallback) gsampler_main_window_connect_proxy_cb, self, 0);
	g_signal_connect_object ((GtkUIManager*) self->priv->_ui_manag, "disconnect-proxy",
	                         (GCallback) gsampler_main_window_disconnect_proxy_cb, self, 0);
	g_signal_connect_object ((GtkUIManager*) self->priv->_ui_manag, "post-activate",
	                         (GCallback) gsampler_main_window_on_action_activated, self, 0);

	main_box = gtk_vbox_new (FALSE, 0);
	self->priv->paned1 = (GtkPaned*) gtk_hpaned_new ();
	gtk_container_set_border_width ((GtkContainer*) self->priv->paned1, 5);
	self->priv->paned2 = (GtkPaned*) gtk_hpaned_new ();
	gtk_paned_set_position (self->priv->paned1, 
	                        gsampler_option_get_integer (self->priv->_option, "General", "paned1"));
	gtk_paned_set_position (self->priv->paned2, 
	                        gsampler_option_get_integer (self->priv->_option, "General", "paned2"));

	sec_box = gtk_vbox_new (FALSE, 0);
	vol_box = gtk_hbox_new (FALSE, 5);						   
	vol_img = gtk_image_new_from_stock (GSAMPLER_STOCK_SOUND, GTK_ICON_SIZE_BUTTON);
	vol_lbl = gtk_label_new (NULL);
	gtk_label_set_markup ((GtkLabel*) vol_lbl, _("<b>Volume :</b>"));
	self->priv->main_vol = (GtkAdjustment*) gtk_adjustment_new (80, 0, 100, 1, 10, 0);
	vol_scale = (GtkScale*) gtk_hscale_new (self->priv->main_vol);
	gtk_scale_set_value_pos (vol_scale, GTK_POS_RIGHT);
	self->priv->master_draw = (GtkDrawingArea*) gtk_drawing_area_new ();
	g_object_set ((GtkWidget*) self->priv->master_draw, "width-request", 50, NULL);
	self->priv->notebook = (GtkNotebook*) gtk_notebook_new ();
	gtk_container_set_border_width ((GtkContainer*) self->priv->notebook, 2);
	
	self->priv->_statusbar = gsampler_status_bar_new ();					   
	self->priv->_keyboard = gsampler_keyboard_new ();
	self->priv->_toolbar = (GtkToolbar*) gtk_ui_manager_get_widget ((GtkUIManager*) self->priv->_ui_manag, "/toolbar");
	self->priv->_menu = (GtkMenuBar*) gtk_ui_manager_get_widget ((GtkUIManager*) self->priv->_ui_manag, "/menubar");
	self->priv->_dev_manag = gsampler_device_manager_get_instance ();
	self->priv->_chan_manag = gsampler_channel_manager_get_instance ();
	self->priv->console = gsampler_console_new ();
	self->priv->_data_base = gsampler_data_base_new ();
	self->priv->_midi_map = gsampler_midi_map_get_instance ();
	self->priv->_session = gsampler_session_new ();
							   
	map_label = gtk_label_new (_("Midi Instrument Map"));
	db_label = gtk_label_new (_("Database"));

	gsampler_ui_manager_group_set_sensitive (self->priv->_ui_manag, "client", FALSE);
	gtk_box_pack_start ((GtkBox*) vol_box, vol_img, FALSE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) vol_box, vol_lbl, FALSE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) vol_box, (GtkWidget*) vol_scale, TRUE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) vol_box, (GtkWidget*) self->priv->master_draw, FALSE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) sec_box, (GtkWidget*) vol_box, FALSE, TRUE, 5);
	gtk_box_pack_start ((GtkBox*) sec_box, (GtkWidget*) self->priv->_chan_manag, TRUE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) sec_box, (GtkWidget*) self->priv->_keyboard, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) sec_box, (GtkWidget*) self->priv->console, FALSE, TRUE, 0);
	gtk_notebook_append_page (self->priv->notebook, (GtkWidget*) self->priv->_data_base, db_label);
	gtk_notebook_set_tab_reorderable (self->priv->notebook, (GtkWidget*) self->priv->_data_base, TRUE);
	gtk_notebook_append_page (self->priv->notebook, (GtkWidget*) self->priv->_midi_map, map_label);
	gtk_notebook_set_tab_reorderable (self->priv->notebook, (GtkWidget*) self->priv->_midi_map, TRUE);	
	gtk_paned_pack1 (self->priv->paned1, (GtkWidget*) self->priv->_dev_manag, TRUE, TRUE);
	gtk_paned_pack2 (self->priv->paned1, (GtkWidget*) self->priv->paned2, TRUE, TRUE);
	gtk_paned_pack1 (self->priv->paned2, (GtkWidget*) sec_box, TRUE, TRUE);
	gtk_paned_pack2 (self->priv->paned2, (GtkWidget*) self->priv->notebook, TRUE, TRUE);
	gtk_box_pack_start ((GtkBox*) main_box, (GtkWidget*) self->priv->_menu, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) main_box, (GtkWidget*) self->priv->_toolbar, FALSE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) main_box, (GtkWidget*) self->priv->paned1, TRUE, TRUE, 0);
	gtk_box_pack_start ((GtkBox*) main_box, (GtkWidget*) self->priv->_statusbar, FALSE, TRUE, 0);
	gtk_container_add ((GtkContainer*) self, (GtkWidget*) main_box);

#if GTK_CHECK_VERSION (2, 90, 0)
	GtkStyleContext* context = gtk_widget_get_style_context ((GtkWidget*) self->priv->_toolbar);
	gtk_style_context_add_class (context, "primary-toolbar");
#endif
	
	gtk_widget_show_all ((GtkWidget*) self);

	gboolean view;
	view = gsampler_option_get_boolean (self->priv->_option, "General", "statusbar");
	if (!view) 
	{
		gtk_widget_hide ((GtkWidget*) self->priv->_statusbar);	
	}
#if GTK_CHECK_VERSION (2, 90, 0)
	gtk_window_set_has_resize_grip ((GtkWindow*) self, FALSE);
#else
	gtk_statusbar_set_has_resize_grip ((GtkStatusbar*) self->priv->_statusbar, FALSE);
#endif	
	gsampler_ui_manager_action_set_toggled (self->priv->_ui_manag, "ViewStatusbar", view);
	view = gsampler_option_get_boolean (self->priv->_option, "General", "toolbar");
	if (!view) 
	{
		gtk_widget_hide ((GtkWidget*) self->priv->_toolbar);
	}
	gsampler_ui_manager_action_set_toggled (self->priv->_ui_manag, "ViewToolbar", view);
	view = gsampler_option_get_boolean (self->priv->_option, "General", "keyboard");
	if (!view) 
	{
		gtk_widget_hide ((GtkWidget*) self->priv->_keyboard);
	}
	gsampler_ui_manager_action_set_toggled (self->priv->_ui_manag, "ViewKeyboard", view);
	view = gsampler_option_get_boolean (self->priv->_option, "General", "console");
	if (!view) 
	{
		gtk_widget_hide ((GtkWidget*) self->priv->console);
	}
	gsampler_ui_manager_action_set_toggled (self->priv->_ui_manag, "ViewConsole", view);


	gtk_notebook_set_current_page (self->priv->notebook, 
	                               gsampler_option_get_integer (self->priv->_option, "General", "tab"));
	
	self->priv->font = g_strdup (gsampler_option_get_string (self->priv->_option, "General", "font"));
	
	gsampler_main_window_update_title (self);

#if GTK_CHECK_VERSION (2, 90, 0)
	g_signal_connect_object ((GtkWidget*) self->priv->master_draw, "draw", 
	                         (GCallback) gsampler_main_window_on_draw_event, self, 0);
#else
	g_signal_connect_object ((GtkWidget*) self->priv->master_draw, "expose_event", 
	                         (GCallback) gsampler_main_window_on_expose_event, self, 0);
#endif
	g_signal_connect_object ((GtkWidget*) self, "delete-event", 
	                         (GCallback) gsampler_main_window_on_quit, NULL, 0);
	return obj;
}

static void 
gsampler_main_window_init (GsamplerMainWindow * self) 
{
	self->priv = GSAMPLER_MAIN_WINDOW_GET_PRIVATE (self);
	self->priv->total_stream = g_strdup ("0");
	self->priv->total_voice = g_strdup ("0");
	self->priv->status = FALSE;
}

static void 
gsampler_main_window_finalize (GObject* obj) 
{
	GsamplerMainWindow * self;
	self = GSAMPLER_MAIN_WINDOW (obj);
	g_object_unref (self->priv->main_vol);
	g_object_unref (self->priv->master_draw);
	g_object_unref (self->priv->paned1);
	g_object_unref (self->priv->paned2);
	g_free (self->priv->font);
	g_free (self->priv->total_stream);
	g_free (self->priv->total_voice);
	g_object_unref (self->priv->_menu);
	g_object_unref (self->priv->_toolbar);
	g_object_unref (self->priv->_statusbar);
	g_object_unref (self->priv->_keyboard);
	g_object_unref (self->priv->_ui_manag);
	g_object_unref (self->priv->console);
	g_object_unref (self->priv->_session);
	g_object_unref (self->priv->_option);
	g_object_unref (self->priv->_dev_manag);
	g_object_unref (self->priv->_chan_manag);
	g_object_unref (self->priv->_midi_map);
	g_object_unref (self->priv->_data_base);
	G_OBJECT_CLASS (gsampler_main_window_parent_class)->finalize (obj);
}

static void 
gsampler_main_window_class_init (GsamplerMainWindowClass * klass) 
{
	gsampler_main_window_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (GsamplerMainWindowPrivate));
	G_OBJECT_CLASS (klass)->finalize = gsampler_main_window_finalize;
	G_OBJECT_CLASS (klass)->constructor = gsampler_main_window_constructor;
}

GsamplerMainWindow* 
gsampler_main_window_new (void) 
{
	return g_object_new (GSAMPLER_TYPE_MAIN_WINDOW, NULL);
}

GsamplerMainWindow* 
gsampler_main_window_get_instance (void) 
{
	if (manwin_instance == NULL) 
	{
		manwin_instance = gsampler_main_window_new ();
	}
	return g_object_ref (manwin_instance);
}

gboolean 
gsampler_main_window_start_client (gpointer data) 
{
	GsamplerClient* client = NULL;
	GsamplerMainWindow* self = GSAMPLER_MAIN_WINDOW (data);
	g_return_val_if_fail (self != NULL, FALSE);

  	if (!g_source_is_destroyed (g_main_current_source ()))
    {
		gsampler_main_window_set_wait_cursor (TRUE);
		client = gsampler_client_get_instance ();
		g_signal_connect_swapped (client, "client-server-start-event", 
			                     (GCallback) gsampler_main_window_set_status_progress, self);
		g_signal_connect_swapped (client, "client-server-event", 
				                  (GCallback) gsampler_console_push_server_event, self->priv->console);
		if (!gsampler_client_start (client)) 
		{
			/*gsampler_main_window_send_message ("Warning", "Can't initialize the sampler\n \
											   If you want gsampler start Linuxsampler set it in the preference" \
			                                   "Click on the restart button or check your configuration");*/
		} else 
		{
			gsampler_main_window_activate_client (self);
			g_signal_connect_swapped (client, "client-echo-event", 
				                  	  (GCallback) gsampler_console_push_event, self->priv->console);
		}
		gsampler_main_window_set_wait_cursor (FALSE);
		g_object_unref (client);
	}
	return FALSE;
}

gdouble 
gsampler_main_window_get_master_volume (void) 
{
	gdouble result = 0.0;

	result = gtk_adjustment_get_value (manwin_instance->priv->main_vol);
	return result;
}

GdkWindow* 
gsampler_main_window_get_gdk_window (void) 
{
	GdkWindow* result = NULL;

	result = g_object_ref (gtk_widget_get_window ((GtkWidget*) manwin_instance));
	return result;
}

void 
gsampler_main_window_set_wait_cursor (gboolean state) 
{
	GdkCursor* watch = NULL;
	GdkWindow* window;
	
	window = gtk_widget_get_window ((GtkWidget*) manwin_instance);
	if (state) {
		watch = gdk_cursor_new (GDK_WATCH);
	}
	gdk_window_set_cursor (window, watch);
}

void 
gsampler_main_window_set_hand_cursor (gboolean state) 
{
	GdkCursor* pointer = NULL;
	GdkCursor* current = NULL;
	GdkWindow* window;
	
	window = gtk_widget_get_window ((GtkWidget*) manwin_instance);
	current = gdk_window_get_cursor (window);
	if (current == NULL || gdk_cursor_get_cursor_type (current) != GDK_WATCH)
	{
		if (state)
			pointer = gdk_cursor_new (GDK_HAND1);
		gdk_window_set_cursor (window, pointer);
	}
}

gboolean
gsampler_man_window_status (void)
{
	return manwin_status;
}
